/*
 * Copyright 2016 SiliconMotion Inc.
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License version 2. See the file COPYING in the main
 * directory of this archive for more details.
 *
 */

#include <linux/types.h>
#include <drm/drm_crtc.h>
#include <drm/drm_modes.h>

#include "ddk768/IT6251.h"
#include "ddk768/ddk768_2d.h"
#include "ddk768/ddk768_cursor.h"
#include "ddk768/ddk768_display.h"
#include "ddk768/ddk768_hdmi.h"
#include "ddk768/ddk768_help.h"
#include "ddk768/ddk768_mode.h"
#include "ddk768/ddk768_power.h"
#include "ddk768/ddk768_reg.h"
#include "ddk768/ddk768_video.h"

#include "smi_dbg.h"

struct smi_768_register {
	uint32_t clock_enable, pll_ctrl[3];

	uint32_t primary_display_ctrl[10], lvds_ctrl, primary_hwcurs_ctrl[4];
	uint32_t secondary_display_ctrl[10], secondary_hwcurs_ctrl[4];
};

int IT6251_initialized = 0;

void hw768_enable_lvds(int channels)
{
	unsigned long value;
	if (channels == 1) {
		pokeRegisterDWord(0x80020, 0x31E30000);
		pokeRegisterDWord(0x8002C, 0x74001200);
	} else {
		pokeRegisterDWord(0x80020, 0x31E3F71D);
		pokeRegisterDWord(0x8002C, 0x750FED02);
		value = peekRegisterDWord(DISPLAY_CTRL);
		value = FIELD_SET(value, DISPLAY_CTRL, LVDS_OUTPUT_FORMAT, CHANNEL0_48BIT);
		value = FIELD_SET(value, DISPLAY_CTRL, PIXEL_CLOCK_SELECT, HALF);
		value = FIELD_SET(value, DISPLAY_CTRL, DOUBLE_PIXEL_CLOCK, ENABLE);
		pokeRegisterDWord(DISPLAY_CTRL, value);

		value = peekRegisterDWord(DISPLAY_CTRL + CHANNEL_OFFSET);
		value = FIELD_SET(value, DISPLAY_CTRL, LVDS_OUTPUT_FORMAT, CHANNEL0_48BIT);
		pokeRegisterDWord(DISPLAY_CTRL + CHANNEL_OFFSET, value);
	}

	//set GPIO 0 to high and PWM 0 to control LVDS dimming for customer Innodisk
	value = peekRegisterDWord(0x10004);
	value |= 1;
	pokeRegisterDWord(0x10004, value);

	value = peekRegisterDWord(0x10000);
	value |= 1;
	pokeRegisterDWord(0x10000, value);

	value = peekRegisterDWord(0x10010);
	value |= 1 << 17;
	pokeRegisterDWord(0x10010, value);

	pokeRegisterDWord(0x10020, 0xfff00001);
}

void hw768_suspend(struct smi_768_register *pSave)
{
	int i;

	pSave->clock_enable = peekRegisterDWord(CLOCK_ENABLE);
	for (i = 0; i < 3; i++)
		pSave->pll_ctrl[i] = peekRegisterDWord(MCLK_PLL + i * 4);

	for (i = 0; i < 10; i++)
		pSave->primary_display_ctrl[i] = peekRegisterDWord(DISPLAY_CTRL + i * 4);
	pSave->lvds_ctrl = peekRegisterDWord(LVDS_CTRL2);
	for (i = 0; i < 4; i++)
		pSave->primary_hwcurs_ctrl[i] = peekRegisterDWord(HWC_CONTROL + i * 4);

	for (i = 0; i < 10; i++)
		pSave->secondary_display_ctrl[i] = peekRegisterDWord(0x8000 + DISPLAY_CTRL + i * 4);
	for (i = 0; i < 4; i++)
		pSave->secondary_hwcurs_ctrl[i] = peekRegisterDWord(0x8000 + HWC_CONTROL + i * 4);
}

void hw768_resume(struct smi_768_register *pSave)
{
	int i;

	pokeRegisterDWord(CLOCK_ENABLE, pSave->clock_enable);
	for (i = 0; i < 3; i++)
		pokeRegisterDWord(MCLK_PLL + i * 4, pSave->pll_ctrl[i]);

	for (i = 0; i < 10; i++)
		pokeRegisterDWord(DISPLAY_CTRL + i * 4, pSave->primary_display_ctrl[i]);
	pokeRegisterDWord(LVDS_CTRL2, pSave->lvds_ctrl);
	for (i = 0; i < 4; i++)
		pokeRegisterDWord(HWC_CONTROL + i * 4, pSave->primary_hwcurs_ctrl[i]);

	for (i = 0; i < 10; i++)
		pokeRegisterDWord(0x8000 + DISPLAY_CTRL + i * 4, pSave->secondary_display_ctrl[i]);
	for (i = 0; i < 4; i++)
		pokeRegisterDWord(0x8000 + HWC_CONTROL + i * 4, pSave->secondary_hwcurs_ctrl[i]);
}

void hw768_set_base(int display, int pitch, int base_addr)
{
	if (display == 0) {
		/* Frame buffer base */
		pokeRegisterDWord((FB_ADDRESS),
				  FIELD_SET(0, FB_ADDRESS, STATUS, PENDING) |
					  FIELD_VALUE(0, FB_ADDRESS, ADDRESS, base_addr));

		/* Pitch value (Hardware people calls it Offset) */
		pokeRegisterDWord((FB_WIDTH), FIELD_VALUE(peekRegisterDWord(FB_WIDTH), FB_WIDTH,
							  OFFSET, pitch));
	} else {
		/* Frame buffer base */
		pokeRegisterDWord((FB_ADDRESS + CHANNEL_OFFSET),
				  FIELD_SET(0, FB_ADDRESS, STATUS, PENDING) |
					  FIELD_VALUE(0, FB_ADDRESS, ADDRESS, base_addr));

		/* Pitch value (Hardware people calls it Offset) */
		pokeRegisterDWord((FB_WIDTH + CHANNEL_OFFSET),
				  FIELD_VALUE(peekRegisterDWord(FB_WIDTH + CHANNEL_OFFSET),
					      FB_WIDTH, OFFSET, pitch));
	}
}

mode_parameter_t convert_drm_mode_to_ddk_mode(struct drm_display_mode mode)
{
	mode_parameter_t modeP;

	modeP.horizontal_total = mode.htotal;
	modeP.horizontal_display_end = mode.hdisplay;
	modeP.horizontal_sync_start = mode.hsync_start;
	modeP.horizontal_sync_width = mode.hsync_end - mode.hsync_start;
	modeP.horizontal_sync_polarity = mode.flags & DRM_MODE_FLAG_PHSYNC ? POS : NEG;

	/* Vertical timing. */
	modeP.vertical_total = mode.vtotal;
	modeP.vertical_display_end = mode.vdisplay;
	modeP.vertical_sync_start = mode.vsync_start;
	modeP.vertical_sync_height = mode.vsync_end - mode.vsync_start;
	modeP.vertical_sync_polarity = mode.flags & DRM_MODE_FLAG_PVSYNC ? POS : NEG;

	/* Refresh timing. */
	modeP.pixel_clock = mode.clock * 1000;
	modeP.horizontal_frequency = 0;
	modeP.vertical_frequency = 0;

	/* Clock Phase. This clock phase only applies to Panel. */
	modeP.clock_phase_polarity = NEG;

	return modeP;
}

struct drm_display_mode *convert_lvds_mode_to_drm_mode(struct drm_device *dev, int index)
{
	struct drm_display_mode *ddm;
	extern mode_parameter_t LVDSModeTbl;

	ddm = drm_mode_create(dev);

	ddm->htotal = LVDSModeTbl.horizontal_total;
	ddm->hdisplay = LVDSModeTbl.horizontal_display_end;
	ddm->hsync_start = LVDSModeTbl.horizontal_sync_start;
	ddm->hsync_end = ddm->hsync_start + LVDSModeTbl.horizontal_sync_width;

	ddm->vtotal = LVDSModeTbl.vertical_total;
	ddm->vdisplay = LVDSModeTbl.vertical_display_end;
	ddm->vsync_start = LVDSModeTbl.vertical_sync_start;
	ddm->vsync_end = ddm->vsync_start + LVDSModeTbl.vertical_sync_height;

	ddm->clock = LVDSModeTbl.pixel_clock / 1000;

	if (LVDSModeTbl.horizontal_sync_polarity == POS)
		ddm->flags |= DRM_MODE_FLAG_PHSYNC;

	if (LVDSModeTbl.vertical_sync_polarity == POS)
		ddm->flags |= DRM_MODE_FLAG_PVSYNC;

	drm_mode_set_name(ddm);
	return ddm;
}

void hw768_init_hdmi(void)
{
	HDMI_Init();
}

bool hw768_init_DP(void)
{
	if (IT6251_probe() == 0) {
		IT6251_Init();
		IT6251_initialized = 1;
		return true;
	} else {
		IT6251_initialized = 0;
		return false;
	}
}

int hw768_set_hdmi_mode(logicalMode_t *pLogicalMode, struct drm_display_mode mode, bool isHDMI)
{
	int ret = 1;
	mode_parameter_t modeParam = { 0 };

	pLogicalMode->hz = pLogicalMode->x == 3840 ? 30 : 60;

	// set HDMI parameters
	HDMI_Disable_Output();
	if (mode.clock != 0)
		modeParam = convert_drm_mode_to_ddk_mode(mode);
	ret = HDMI_Set_Mode(pLogicalMode, &modeParam, isHDMI);
	return ret;
}

int hw768_en_dis_interrupt(int status, int pipe)
{
	if (status == 0) {
		pokeRegisterDWord(INT_MASK,
				  (pipe == CHANNEL0_CTRL) ?
					  FIELD_SET(0, INT_MASK, CHANNEL1_VSYNC, DISABLE) :
					  FIELD_SET(0, INT_MASK, CHANNEL0_VSYNC, DISABLE));
	} else {
		pokeRegisterDWord(INT_MASK, (pipe == CHANNEL1_CTRL) ?
						    FIELD_SET(0, INT_MASK, CHANNEL1_VSYNC, ENABLE) :
						    FIELD_SET(0, INT_MASK, CHANNEL0_VSYNC, ENABLE));
	}
	return 0;
}
void hw768_HDMI_Enable_Output(void)
{
	HDMI_Enable_Output();
}

void hw768_HDMI_Disable_Output(void)
{
	HDMI_Disable_Output();
}

int hw768_get_hdmi_edid(unsigned char *pEDIDBuffer)
{
	int ret;
	enableHdmI2C(1);
	ret = HDMI_Read_Edid(pEDIDBuffer, 128);
	enableHdmI2C(0);

	return ret;
}

// EDID is assumed to be read 256 bytes.
int hw768_get_dp_edid(unsigned char *pEDIDBuffer)
{
	return IT6251_Read_EDID(pEDIDBuffer);
}

int hw768_check_dp_status(void)
{
	return IT6251_Check();
}

int hw768_check_iis_interrupt(void)
{
	unsigned long value;

	value = peekRegisterDWord(INT_STATUS);

	if (FIELD_GET(value, INT_STATUS, I2S) == INT_STATUS_I2S_ACTIVE)
		return true;
	else
		return false;
}

int hw768_check_vsync_interrupt(int path)
{
	unsigned long value1, value2;

	value1 = peekRegisterDWord(RAW_INT);
	value2 = peekRegisterDWord(INT_MASK);

	if (path == CHANNEL0_CTRL) {
		if ((FIELD_GET(value1, RAW_INT, CHANNEL0_VSYNC) == RAW_INT_CHANNEL0_VSYNC_ACTIVE) &&
		    (FIELD_GET(value2, INT_MASK, CHANNEL0_VSYNC) ==
		     INT_MASK_CHANNEL0_VSYNC_ENABLE)) {
			return true;
		}
	} else {
		if ((FIELD_GET(value1, RAW_INT, CHANNEL1_VSYNC) == RAW_INT_CHANNEL1_VSYNC_ACTIVE) &&
		    (FIELD_GET(value2, INT_MASK, CHANNEL1_VSYNC) ==
		     INT_MASK_CHANNEL1_VSYNC_ENABLE)) {
			return true;
		}
	}

	return false;
}

void hw768_clear_vsync_interrupt(int path)
{
	unsigned long value;

	value = peekRegisterDWord(RAW_INT);

	if (path == CHANNEL0_CTRL) {
		pokeRegisterDWord(RAW_INT, FIELD_SET(value, RAW_INT, CHANNEL0_VSYNC, CLEAR));

	} else {
		pokeRegisterDWord(RAW_INT, FIELD_SET(value, RAW_INT, CHANNEL1_VSYNC, CLEAR));
	}
}
long hw768_setMode(logicalMode_t *pLogicalMode, struct drm_display_mode mode)
{
	if (mode.clock == 0 || (pLogicalMode->x == 3840))
		return ddk768_setMode(pLogicalMode);
	else {
		mode_parameter_t ModeParam = convert_drm_mode_to_ddk_mode(mode);
		return ddk768_setCustomMode(pLogicalMode, &ModeParam);
	}
}

int hdmi_int_status = 0;

inline int hdmi_hotplug_detect(void)
{
	int ret = 0;
	unsigned int intMask = peekRegisterDWord(INT_MASK);
	intMask = FIELD_SET(intMask, INT_MASK, HDMI, ENABLE);
	pokeRegisterDWord(INT_MASK, intMask);

	ret = hdmi_detect();

	if (ret == 1) {
		hdmi_int_status = 1;
	} else if (ret == 0) {
		hdmi_int_status = 0;
	} else {
		hdmi_int_status = hdmi_int_status & ret;
	}

	intMask = peekRegisterDWord(INT_MASK);
	intMask = FIELD_SET(intMask, INT_MASK, HDMI, DISABLE);
	pokeRegisterDWord(INT_MASK, intMask);

	return hdmi_int_status;
}

void ddk768_disable_IntMask(void)
{
	pokeRegisterDWord(INT_MASK, 0);
}
