/*
 * Copyright 2016 SiliconMotion Inc.
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License version 2. See the file COPYING in the main
 * directory of this archive for more details.
 *
 */
#include "smi_drv.h"
#if RHEL_VERSION_LOWER_THAN(8, 2)
#include <drm/drmP.h>
#endif
#include "hw750.h"
#include "hw768.h"

extern disp_control_t get_ctrl_from_crtc(struct drm_crtc *);

__attribute__((unused)) static void colorcur2monocur(void *dst, const void *src)
{
	unsigned int *col = (unsigned int *)src;
	unsigned char *mono = (unsigned char *)dst;
	unsigned char pixel = 0;
	char bit_values;

	int i;
	for (i = 0; i < 64 * 64; i++) {
		if (*col >> 24 < 0xe0) {
			bit_values = 0;
		} else {
			int val = *col & 0xff;

			if (val < 0x80) {
				bit_values = 1;
			} else {
				bit_values = 2;
			}
		}
		col++;
		/* Copy bits into cursor byte */
		switch (i & 3) {
		case 0:
			pixel = bit_values;
			break;

		case 1:
			pixel |= bit_values << 2;
			break;

		case 2:
			pixel |= bit_values << 4;
			break;

		case 3:
			pixel |= bit_values << 6;
			*mono = pixel;
			mono++;
			pixel = 0;
			break;
		}
	}
}

#if (RHEL_VERSION_HIGHER_THAN(7, 7) && RHEL_VERSION_LOWER_THAN(8, 0)) ||                           \
	RHEL_VERSION_HIGHER_THAN(8, 1)
static int smi_cursor_update(struct smi_device *sdev, void *dst, void *src, int width, int height)
{
	if (g_specId == SPC_SM768) {
		memcpy_toio(dst, src, width * height * 4);
	} else if (g_specId == SPC_SM750) {
		colorcur2monocur(dst, src);
	}

	return (0);
}

static void smi_hide_cursor(struct smi_device *sdev, disp_control_t disp_ctrl)
{
	if (g_specId == SPC_SM768) {
		ddk768_enableCursor(disp_ctrl, 0);
	} else if (g_specId == SPC_SM750) {
		ddk750_enableCursor(disp_ctrl, 0);
	}
}

static int smi_show_cursor(struct smi_device *sdev, void *src, disp_control_t disp_ctrl,
			   unsigned int width, unsigned int height)
{
#if RHEL_VERSION_HIGHER_THAN(8, 2)
	struct drm_gem_vram_object *cursor_bo;
	s64 off;
#else
	struct smi_bo *cursor_bo = NULL;
	u64 off;
#endif
	void *dst = NULL;
	int ret;

	ENTER();
	cursor_bo = sdev->cursor.cursors[sdev->cursor.next_index];
	if (!cursor_bo) {
		smi_hide_cursor(sdev, disp_ctrl);
		LEAVE(-ENOTSUPP);
	}
#if RHEL_VERSION_HIGHER_THAN(8, 2)
	dst = drm_gem_vram_kmap(cursor_bo, true, NULL);
	if (IS_ERR(dst)) {
		ret = PTR_ERR(dst);
		DRM_ERROR("failed to map cursor updates");
		LEAVE(ret);
	}
	off = drm_gem_vram_offset(cursor_bo);
	if (off < 0) {
		ret = (int)off;
		DRM_ERROR("failed to get cursor scanout address");
		goto err_drm_gem_vram_kunmap;
	}
#else
	ret = smi_bo_reserve(cursor_bo, true);
	if (ret) {
		LEAVE(ret);
	}
	ret = smi_bo_pin(cursor_bo, TTM_PL_FLAG_VRAM, &off);
	if (ret) {
		goto err_drm_unreserve;
	}
	ret = ttm_bo_kmap(&cursor_bo->bo, 0, cursor_bo->bo.num_pages, &cursor_bo->kmap);
	if (ret) {
		goto err_drm_unpin;
	}
	dst = cursor_bo->kmap.virtual;
#endif
	ret = smi_cursor_update(sdev, dst, src, width, height);
	if (ret) {
#if RHEL_VERSION_HIGHER_THAN(8, 2)
		goto err_drm_gem_vram_kunmap;
#else
		goto err_ttm_bo_kunmap;
#endif
	}

	if (g_specId == SPC_SM768) {
		ddk768_initCursor(disp_ctrl, (u32)off, BPP32_BLACK, BPP32_WHITE, BPP32_BLUE);
		ddk768_enableCursor(disp_ctrl, 3);
	} else if (g_specId == SPC_SM750) {
		ddk750_initCursor(disp_ctrl, (u32)off, BPP16_BLACK, BPP16_WHITE, BPP16_BLUE);
		ddk750_enableCursor(disp_ctrl, 1);
	}

#if RHEL_VERSION_HIGHER_THAN(8, 2)
	drm_gem_vram_kunmap(cursor_bo);
#else
	ttm_bo_kunmap(&cursor_bo->kmap);
	smi_bo_unpin(cursor_bo);
	smi_bo_unreserve(cursor_bo);
#endif
	++sdev->cursor.next_index;
	sdev->cursor.next_index %= ARRAY_SIZE(sdev->cursor.cursors);

	LEAVE(0);

#if RHEL_VERSION_HIGHER_THAN(8, 2)
err_drm_gem_vram_kunmap:
	drm_gem_vram_kunmap(cursor_bo);
#else
err_ttm_bo_kunmap:
	ttm_bo_kunmap(&cursor_bo->kmap);
err_drm_unpin:
	smi_bo_unpin(cursor_bo);
err_drm_unreserve:
	smi_bo_unreserve(cursor_bo);
#endif
	LEAVE(ret);
}

int smi_crtc_cursor_set(struct drm_crtc *crtc, struct drm_file *file_priv, unsigned int handle,
			unsigned int width, unsigned int height)
{
	struct drm_device *dev = crtc->dev;
	struct smi_device *sdev = (struct smi_device *)dev->dev_private;
	int next_index = 0;
	void *src = NULL;
#if RHEL_VERSION_HIGHER_THAN(8, 2)
	struct drm_gem_vram_object *cursor_bo = NULL;
#else
	struct smi_bo *cursor_bo = NULL;
#endif
	struct drm_gem_object *obj;
	int ret;
	disp_control_t disp_ctrl = get_ctrl_from_crtc(crtc);

	ENTER();

	if (!handle || !file_priv) {
		smi_hide_cursor(sdev, disp_ctrl);
		LEAVE(0);
	}

	if (width != 64 || height != 64) {
		if (g_specId == SPC_SM768) {
			ddk768_enableCursor(disp_ctrl, 0);
		} else if (g_specId == SPC_SM750) {
			ddk750_enableCursor(disp_ctrl, 0);
		}
		DRM_WARN("cursor too big");
		LEAVE(-EINVAL);
	}

	if (sdev->cursor.next_index == 0)
		next_index = 1;
	else
		next_index = 0;

	obj = drm_gem_object_lookup(file_priv, handle);
	if (!obj)
		LEAVE(-ENOENT);
#if RHEL_VERSION_HIGHER_THAN(8, 2)
	cursor_bo = drm_gem_vram_of_gem(obj);
	ret = drm_gem_vram_pin(cursor_bo, DRM_GEM_VRAM_PL_FLAG_SYSTEM);
	if (ret) {
		DRM_ERROR("failed to pin user buffer");
		goto err_drm_gem_object_put_unlocked;
	}
	src = drm_gem_vram_kmap(cursor_bo, true, NULL);
	if (IS_ERR(src)) {
		ret = PTR_ERR(src);
		DRM_ERROR("failed to kmap user buffer");
		goto err_drm_vram_unpin;
	}
#else
	cursor_bo = gem_to_smi_bo(obj);
	ret = smi_bo_reserve(cursor_bo, true);
	if (ret) {
		DRM_ERROR("failed to reserve user buffer");
		goto err_drm_gem_object_put_unlocked;
	}
	ret = smi_bo_pin(cursor_bo, 0, NULL);
	if (ret) {
		DRM_ERROR("failed to lock user buffer");
		goto err_drm_unreserve;
	}
	ret = ttm_bo_kmap(&cursor_bo->bo, 0, cursor_bo->bo.num_pages, &cursor_bo->kmap);
	if (ret) {
		DRM_ERROR("failed to kmap user buffer");
		goto err_drm_unpin;
	}
	src = cursor_bo->kmap.virtual;
#endif

	ret = smi_show_cursor(sdev, src, disp_ctrl, width, height);
	if (ret)
#if RHEL_VERSION_HIGHER_THAN(8, 2)
		goto err_drm_gem_vram_kunmap;

	drm_gem_vram_kunmap(cursor_bo);
	drm_gem_vram_unpin(cursor_bo);
	drm_gem_object_put_unlocked(obj);
#else
		goto err_drm_kunmap;

	ttm_bo_kunmap(&cursor_bo->kmap);
	smi_bo_unpin(cursor_bo);
	smi_bo_unreserve(cursor_bo);
	drm_gem_object_put_unlocked(obj);
#endif
	LEAVE(0);

#if RHEL_VERSION_HIGHER_THAN(8, 2)
err_drm_gem_vram_kunmap:
	drm_gem_vram_kunmap(cursor_bo);
err_drm_vram_unpin:
	drm_gem_vram_unpin(cursor_bo);
#else
err_drm_kunmap:
	ttm_bo_kunmap(&cursor_bo->kmap);
err_drm_unpin:
	smi_bo_unpin(cursor_bo);
err_drm_unreserve:
	smi_bo_unreserve(cursor_bo);
#endif
err_drm_gem_object_put_unlocked:
	drm_gem_object_put_unlocked(obj);
	LEAVE(ret);
}

int smi_crtc_cursor_move(struct drm_crtc *crtc, int x, int y)
{
	disp_control_t disp_ctrl = get_ctrl_from_crtc(crtc);

	if (g_specId == SPC_SM768) {
		ddk768_setCursorPosition(disp_ctrl, x < 0 ? -x : x, y < 0 ? -y : y, y < 0 ? 1 : 0,
					 x < 0 ? 1 : 0);
	} else if (g_specId == SPC_SM750) {
		ddk750_setCursorPosition(disp_ctrl, x < 0 ? -x : x, y < 0 ? -y : y, y < 0 ? 1 : 0,
					 x < 0 ? 1 : 0);
	}
	return (0);
}

int smi_cursor_init(struct smi_device *sdev)
{
	struct drm_device *dev = sdev->dev;
	int ret;
	int i = 0;
#if RHEL_VERSION_HIGHER_THAN(8, 2)
	struct drm_gem_vram_object *cursorbo;
#else
	struct smi_bo *cursorbo;
#endif

	for (i = 0; i < ARRAY_SIZE(sdev->cursor.cursors); i++) {
#if RHEL_VERSION_HIGHER_THAN(8, 2)
		unsigned long flags = DRM_GEM_VRAM_PL_FLAG_VRAM;
#if RHEL_VERSION_HIGHER_THAN(8, 3)
		flags |= DRM_GEM_VRAM_PL_FLAG_TOPDOWN;
#endif
		cursorbo = drm_gem_vram_create(dev, &dev->vram_mm->bdev,
					       roundup(64 * 64 * 4, PAGE_SIZE), 0, false);
		if (IS_ERR(cursorbo)) {
			ret = PTR_ERR(cursorbo);
			goto err_drm_gem_vram_put;
		}
		ret = drm_gem_vram_pin(cursorbo, flags);
		if (ret) {
			drm_gem_vram_put(cursorbo);
			goto err_drm_gem_vram_put;
		}
#else
		ret = smi_bo_create(dev, roundup(64 * 64 * 4, PAGE_SIZE), 0, 0, NULL, NULL,
				    &cursorbo);
		if (ret)
			goto err_bo_unref;
		ret = smi_bo_reserve(cursorbo, true);
		if (ret) {
			DRM_ERROR("failed to reserve user buffer");
			goto err_bo_unref;
		}
		ret = smi_bo_pin(cursorbo, TTM_PL_FLAG_VRAM, NULL);
		if (ret) {
			DRM_ERROR("failed to lock user buffer");
			smi_bo_unreserve(cursorbo);
			cursorbo = NULL;
			goto err_bo_unref;
		}
		smi_bo_unreserve(cursorbo);
#endif
		sdev->cursor.cursors[i] = cursorbo;
	}
	return 0;
#if RHEL_VERSION_HIGHER_THAN(8, 2)
err_drm_gem_vram_put:
	while (i) {
		--i;
		cursorbo = sdev->cursor.cursors[i];
		drm_gem_vram_unpin(cursorbo);
		drm_gem_vram_put(cursorbo);
		sdev->cursor.cursors[i] = NULL;
	}
#else
err_bo_unref:
	while (i) {
		--i;
		if (sdev->cursor.cursors[i]->pin_count > 0) {
			smi_bo_reserve(sdev->cursor.cursors[i], true);
			smi_bo_unpin(sdev->cursor.cursors[i]);
			smi_bo_unreserve(sdev->cursor.cursors[i]);
		}

		sdev->cursor.cursors[i] = NULL;
	}
#endif
	return ret;
}

void smi_cursor_fini(struct smi_device *sdev)
{
	int i;
#if RHEL_VERSION_HIGHER_THAN(8, 2)
	for (i = 0; i < ARRAY_SIZE(sdev->cursor.cursors); i++) {
		drm_gem_vram_unpin(sdev->cursor.cursors[i]);
		drm_gem_vram_put(sdev->cursor.cursors[i]);
	}
#else
	for (i = 0; i < ARRAY_SIZE(sdev->cursor.cursors); i++) {
		if (sdev->cursor.cursors[i]->pin_count > 0) {
			smi_bo_reserve(sdev->cursor.cursors[i], true);
			smi_bo_unpin(sdev->cursor.cursors[i]);
			smi_bo_unreserve(sdev->cursor.cursors[i]);
		}
		sdev->cursor.cursors[i] = NULL;
	}
#endif
}
#endif
