/*
 * Copyright 2016 SiliconMotion Inc.
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License version 2. See the file COPYING in the main
 * directory of this archive for more details.
 *
 */
#include "smi_drv.h"
#include <linux/console.h>
#include <linux/module.h>

#if RHEL_VERSION_LOWER_THAN(8, 2)
#include <drm/drmP.h>
#else
#include <drm/drm_drv.h>
#include <drm/drm_file.h>
#include <drm/drm_ioctl.h>
#include <drm/drm_print.h>
#if RHEL_VERSION_LOWER_THAN(8, 3)
#include <drm/drm_pci.h>
#else
#include <linux/pci.h>
#endif
#include <drm/drm_vblank.h>
#endif

#if (RHEL_VERSION_HIGHER_THAN(7, 9) && RHEL_VERSION_LOWER_THAN(8, 0)) ||                           \
	RHEL_VERSION_HIGHER_THAN(8, 1)
#include <drm/drm_probe_helper.h>
#endif
#include <drm/drm_crtc_helper.h>

#include "hw750.h"
#include "hw768.h"

#include "sm768_lvds.h"

int smi_modeset = -1;
int smi_indent = 0;
int smi_bpp = 32;
int force_connect = 0;
int g_specId = 0;
int smi_pat = 0xff;
int lvds_channel = 0;

module_param(smi_pat, int, S_IWUSR | S_IRUSR);

MODULE_PARM_DESC(modeset, "Disable/Enable modesetting");
module_param_named(modeset, smi_modeset, int, 0400);
MODULE_PARM_DESC(bpp, "Max bits-per-pixel (default:16)");
module_param_named(bpp, smi_bpp, int, 0400);
MODULE_PARM_DESC(nopnp, "Force conncet to the monitor without monitor EDID (default:0)");
module_param_named(nopnp, force_connect, int, 0400);
MODULE_PARM_DESC(lvds, "LVDS Channel, 0=disable 1=single channel, 2=dual channel (default:0)");
module_param_named(lvds, lvds_channel, int, 0400);

/*
 * This is the generic driver code. This binds the driver to the drm core,
 * which then performs further device association and calls our graphics init
 * functions
 */
#define PCI_VENDOR_ID_SMI 0x126f
#define PCI_DEVID_SM750 0x0750
#define PCI_DEVID_SM768 0x0768

static struct drm_driver driver;

/* only bind to the smi chip in qemu */
static const struct pci_device_id pciidlist[] = {
	{ PCI_VENDOR_ID_SMI, PCI_DEVID_SM750, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ PCI_VENDOR_ID_SMI, PCI_DEVID_SM768, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{
		0,
	}
};

static void claim(void)
{
	printk("+-------------SMI Driver Information------------+\n");
	printk("Release type: " RELEASE_TYPE "\n");
	printk("Driver version: v" _version_ "\n");
	printk("Support products: " SUPPORT_CHIP "\n");
	printk("+-----------------------------------------------+\n");
}

static int smi_drm_freeze(struct drm_device *dev);
#if RHEL_VERSION_HIGHER_THAN(8, 2)
static int smi_drm_resume(struct drm_device *dev, bool thaw);
#else
static int smi_drm_resume(struct drm_device *dev);
static int smi_drm_thaw(struct drm_device *dev);
#endif

#if (RHEL_VERSION_HIGHER_THAN(8, 0) && RHEL_VERSION_LOWER_THAN(8, 1)) ||                           \
	(RHEL_VERSION_HIGHER_THAN(7, 3) && RHEL_VERSION_LOWER_THAN(7, 7))
static int smi_kick_out_firmware_fb(struct pci_dev *pdev)
{
	struct apertures_struct *ap;
	bool primary = false;

	ap = alloc_apertures(1);
	if (!ap)
		return -ENOMEM;

	ap->ranges[0].base = pci_resource_start(pdev, 0);
	ap->ranges[0].size = pci_resource_len(pdev, 0);

#ifdef CONFIG_X86
	primary = pdev->resource[PCI_ROM_RESOURCE].flags & IORESOURCE_ROM_SHADOW;
#endif

#if RHEL_VERSION_HIGHER_THAN(7, 4)
	drm_fb_helper_remove_conflicting_framebuffers(ap, "smidrmfb", primary);
#else
	remove_conflicting_framebuffers(ap, "smidrmfb", primary);
#endif

	kfree(ap);

	return 0;
}
#endif

static int smi_pci_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
#if RHEL_VERSION_HIGHER_THAN(8, 3)
	struct drm_device *dev;
#endif
	int ret __attribute__((unused)) = 0;

	claim();
	if (ent->vendor != PCI_VENDOR_ID_SMI &&
	    !(ent->device == PCI_DEVID_LYNX_EXP || ent->device == PCI_DEVID_SM768)) {
		return -ENODEV;
	}

#if RHEL_VERSION_HIGHER_THAN(8, 3)
	drm_fb_helper_remove_conflicting_pci_framebuffers(pdev, "smidrmfb");
#elif (RHEL_VERSION_HIGHER_THAN(7, 7) && RHEL_VERSION_LOWER_THAN(8, 0)) ||                         \
	RHEL_VERSION_HIGHER_THAN(8, 1)
	drm_fb_helper_remove_conflicting_pci_framebuffers(pdev, 0, "smidrmfb");
#else
	ret = smi_kick_out_firmware_fb(pdev);
	if (ret) {
		return ret;
	}
#endif

	if (ent->vendor == PCI_VENDOR_ID_SMI && ent->device == PCI_DEVID_SM768) {
		if (ReadLVDSConfiguration() < 0) {
			dbg_msg("ReadLVDSConfiguration failed!");
		}

		if (lvds_channel != 0) {
			dbg_msg("LVDS channel set to %d\n", lvds_channel);
		}
	}

#if RHEL_VERSION_HIGHER_THAN(8, 3)
	ret = pci_enable_device(pdev);
	if (ret)
		return ret;

	dev = drm_dev_alloc(&driver, &pdev->dev);
	if (IS_ERR(dev)) {
		ret = PTR_ERR(dev);
		goto err_pci_disable_device;
	}

	dev->pdev = pdev;
	pci_set_drvdata(pdev, dev);

	ret = smi_driver_load(dev, ent->driver_data);
	if (ret)
		goto err_drm_dev_put;

	ret = drm_dev_register(dev, ent->driver_data);
	if (ret)
		goto err_smi_driver_unload;

	return 0;

err_smi_driver_unload:
	smi_driver_unload(dev);
err_drm_dev_put:
	drm_dev_put(dev);
err_pci_disable_device:
	pci_disable_device(pdev);
	return ret;
#else
	return drm_get_pci_dev(pdev, ent, &driver);
#endif
}

static void smi_pci_remove(struct pci_dev *pdev)
{
	struct drm_device *dev = pci_get_drvdata(pdev);

#if RHEL_VERSION_HIGHER_THAN(8, 3)
	drm_dev_unregister(dev);
	smi_driver_unload(dev);
	drm_dev_put(dev);
#else
	drm_put_dev(dev);
#endif
}

#if RHEL_VERSION_HIGHER_THAN(8, 2)
static int smi_drm_freeze(struct drm_device *dev)
{
	struct smi_device *sdev = dev->dev_private;
	ENTER();

	drm_kms_helper_poll_disable(dev);
	drm_fb_helper_set_suspend_unlocked(dev->fb_helper, 1);

	if (g_specId == SPC_SM750)
		hw750_suspend(sdev->regsave);
	else if (g_specId == SPC_SM768)
		hw768_suspend(sdev->regsave_768);

	pci_save_state(dev->pdev);

	LEAVE(0);
}

static int smi_drm_resume(struct drm_device *dev, bool thaw)
{
	struct smi_device *sdev = dev->dev_private;
	ENTER();

	if (!thaw) {
		pci_set_power_state(dev->pdev, PCI_D0);
		pci_restore_state(dev->pdev);
		if (pci_enable_device(dev->pdev)) {
			LEAVE(-EIO);
		}
	}

	if (g_specId == SPC_SM750)
		hw750_resume(sdev->regsave);
	else if (g_specId == SPC_SM768) {
		hw768_resume(sdev->regsave_768);
		if (lvds_channel == 1)
			hw768_enable_lvds(1);
		else if (lvds_channel == 2) {
			hw768_enable_lvds(2);
			EnableDoublePixel(0);
		}
	}

	drm_fb_helper_set_suspend_unlocked(dev->fb_helper, 0);
	drm_kms_helper_poll_enable(dev);

	LEAVE(0);
}
#else
static int smi_drm_freeze(struct drm_device *dev)
{
	struct smi_device *sdev = dev->dev_private;
	ENTER();

	drm_kms_helper_poll_disable(dev);

	if (sdev->mode_info.gfbdev) {
		console_lock();
#if RHEL_VERSION_HIGHER_THAN(7, 4)
		drm_fb_helper_set_suspend(&sdev->mode_info.gfbdev->helper, 1);
#else
		fb_set_suspend(sdev->mode_info.gfbdev->helper.fbdev, 1);
#endif
		console_unlock();
	}

	if (g_specId == SPC_SM750)
		hw750_suspend(sdev->regsave);
	else if (g_specId == SPC_SM768)
		hw768_suspend(sdev->regsave_768);

	pci_save_state(dev->pdev);

	LEAVE(0);
}

static int smi_drm_resume(struct drm_device *dev)
{
	struct smi_device *UNUSED(sdev) = dev->dev_private;
	int ret;
	ENTER();

	if (pci_enable_device(dev->pdev))
		return -EIO;

	ret = smi_drm_thaw(dev);
	if (ret)
		return ret;

	drm_kms_helper_poll_enable(dev);

	LEAVE(0);
}

static int smi_drm_thaw(struct drm_device *dev)
{
	struct smi_device *sdev = dev->dev_private;
	ENTER();

	drm_mode_config_reset(dev);
	drm_helper_resume_force_mode(dev);

	if (sdev->mode_info.gfbdev) {
		console_lock();
#if RHEL_VERSION_HIGHER_THAN(7, 4)
		drm_fb_helper_set_suspend(&sdev->mode_info.gfbdev->helper, 0);
#else
		fb_set_suspend(sdev->mode_info.gfbdev->helper.fbdev, 0);
#endif
#if RHEL_VERSION_LOWER_THAN(8, 2)
		smi_fb_zfill(dev, sdev->mode_info.gfbdev);
#endif

		console_unlock();
	}

	if (g_specId == SPC_SM750)
		hw750_resume(sdev->regsave);
	else if (g_specId == SPC_SM768) {
		hw768_resume(sdev->regsave_768);
		if (lvds_channel == 1)
			hw768_enable_lvds(1);
		else if (lvds_channel == 2) {
			hw768_enable_lvds(2);
			EnableDoublePixel(0);
		}
	}

	LEAVE(0);
}
#endif

static int smi_pm_suspend(struct device *dev)
{
	struct pci_dev *pdev = to_pci_dev(dev);
	struct drm_device *ddev = pci_get_drvdata(pdev);
	int error;

	error = smi_drm_freeze(ddev);
	if (error)
		return error;

	pci_disable_device(pdev);
	pci_set_power_state(pdev, PCI_D3hot);
	return 0;
}

static int smi_pm_resume(struct device *dev)
{
	struct drm_device *ddev = dev_get_drvdata(dev);

#if RHEL_VERSION_HIGHER_THAN(8, 2)
	return smi_drm_resume(ddev, false);
#else
	return smi_drm_resume(ddev);
#endif
}

static int smi_pm_freeze(struct device *dev)
{
	struct drm_device *ddev = dev_get_drvdata(dev);

	return smi_drm_freeze(ddev);
}

static int smi_pm_thaw(struct device *dev)
{
	struct drm_device *ddev = dev_get_drvdata(dev);
#if RHEL_VERSION_HIGHER_THAN(8, 2)
	return smi_drm_resume(ddev, true);
#else
	return smi_drm_thaw(ddev);
#endif
}

static int smi_pm_poweroff(struct device *dev)
{
	struct drm_device *ddev = dev_get_drvdata(dev);

	return smi_drm_freeze(ddev);
}

static int smi_enable_vblank(struct drm_device *dev, unsigned int pipe)
{
	if (g_specId == SPC_SM750) {
		hw750_en_dis_interrupt(1, pipe);
	} else if (g_specId == SPC_SM768) {
		hw768_en_dis_interrupt(1, pipe);
	}
	return 0;
}

static void smi_disable_vblank(struct drm_device *dev, unsigned int pipe)
{
	if (g_specId == SPC_SM750) {
		hw750_en_dis_interrupt(0, pipe);
	} else if (g_specId == SPC_SM768) {
		hw768_en_dis_interrupt(0, pipe);
	}
}

static void smi_irq_preinstall(struct drm_device *dev)
{
	// To Do....
	/* Disable *all* interrupts */

	/* Clear bits if they're already high */
}

static int smi_irq_postinstall(struct drm_device *dev)
{
	return 0;
}

static void smi_irq_uninstall(struct drm_device *dev)
{
	/* Disable *all* interrupts */
	if (g_specId == SPC_SM750) {
		ddk750_disable_IntMask();
	} else if (g_specId == SPC_SM768) {
		ddk768_disable_IntMask();
	}
}

#ifndef DRM_IRQ_ARGS
#define DRM_IRQ_ARGS int irq, void *arg
#endif
irqreturn_t smi_drm_interrupt(DRM_IRQ_ARGS)
{
	struct drm_device *dev = (struct drm_device *)arg;

	int handled = 0;

	if (g_specId == SPC_SM750) {
		if (hw750_check_vsync_interrupt(0)) {
			/* Clear the panel VSync Interrupt */
			drm_handle_vblank(dev, 0);
			handled = 1;
			hw750_clear_vsync_interrupt(0);
		}
		if (hw750_check_vsync_interrupt(1)) {
			drm_handle_vblank(dev, 1);
			handled = 1;
			hw750_clear_vsync_interrupt(1);
		}
	} else if (g_specId == SPC_SM768) {
		if (hw768_check_vsync_interrupt(0)) {
			/* Clear the panel VSync Interrupt */
			drm_handle_vblank(dev, 0);
			handled = 1;
			hw768_clear_vsync_interrupt(0);
		}
		if (hw768_check_vsync_interrupt(1)) {
			drm_handle_vblank(dev, 1);
			handled = 1;
			hw768_clear_vsync_interrupt(1);
		}
	}

	if (handled)
		return IRQ_HANDLED;
	return IRQ_NONE;
}

static const struct dev_pm_ops smi_pm_ops = {
	.suspend = smi_pm_suspend,
	.resume = smi_pm_resume,
	.freeze = smi_pm_freeze,
	.thaw = smi_pm_thaw,
	.poweroff = smi_pm_poweroff,
	.restore = smi_pm_resume,
};

static struct pci_driver smi_pci_driver = {
	.name = DRIVER_NAME,
	.id_table = pciidlist,
	.probe = smi_pci_probe,
	.remove = smi_pci_remove,
	.driver.pm = &smi_pm_ops,
};

#if RHEL_VERSION_HIGHER_THAN(8, 3)
DEFINE_DRM_GEM_FOPS(smi_driver_fops);
#else
static const struct file_operations smi_driver_fops = {
#if RHEL_VERSION_HIGHER_THAN(8, 2)
	DRM_VRAM_MM_FILE_OPERATIONS
#else
	.open = drm_open,
	.release = drm_release,
	.unlocked_ioctl = drm_ioctl,
	.mmap = smi_mmap,
	.poll = drm_poll,
#ifdef CONFIG_COMPAT
	.compat_ioctl = drm_compat_ioctl,
#endif
	.read = drm_read,
	.llseek = no_llseek,
#endif
	.owner = THIS_MODULE,
};
#endif

static struct drm_driver driver = {
	.driver_features = DRIVER_HAVE_IRQ | DRIVER_GEM | DRIVER_MODESET
#if defined(PRIME) && RHEL_VERSION_LOWER_THAN(8, 3)
			   | DRIVER_PRIME
#endif
#if RHEL_VERSION_LOWER_THAN(7, 9) ||                                                               \
	(RHEL_VERSION_HIGHER_THAN(8, 0) && RHEL_VERSION_LOWER_THAN(8, 1))
			   | DRIVER_IRQ_SHARED
#endif
	,
#if RHEL_VERSION_LOWER_THAN(8, 3)
	.load = smi_driver_load,
	.unload = smi_driver_unload,
#endif
#if RHEL_VERSION_LOWER_THAN(7, 5)
	.set_busid = drm_pci_set_busid,
#endif
	.fops = &smi_driver_fops,
	.name = DRIVER_NAME,
	.desc = DRIVER_DESC,
	.date = DRIVER_DATE,
	.major = DRIVER_MAJOR,
	.minor = DRIVER_MINOR,
	.patchlevel = DRIVER_PATCHLEVEL,
#if RHEL_VERSION_LOWER_THAN(7, 3)
	.gem_init_object = smi_gem_init_object,
#endif
	.dumb_create = smi_dumb_create,
#if RHEL_VERSION_HIGHER_THAN(8, 2)
	.dumb_map_offset = drm_gem_vram_driver_dumb_mmap_offset,
	.gem_prime_mmap = drm_gem_prime_mmap,
#else
#if RHEL_VERSION_LOWER_THAN(7, 4)
	.gem_free_object = smi_gem_free_object,
#else
	.gem_free_object_unlocked = smi_gem_free_object,
#endif
	.dumb_map_offset = smi_dumb_mmap_offset,
#if RHEL_VERSION_LOWER_THAN(7, 4)
	.dumb_destroy = smi_dumb_destroy,
#else
	.dumb_destroy = drm_gem_dumb_destroy,
#endif
#endif
#if (RHEL_VERSION_HIGHER_THAN(7, 4) && RHEL_VERSION_LOWER_THAN(7, 5))
	.get_vblank_counter = drm_vblank_no_hw_counter,
#endif
	.enable_vblank = smi_enable_vblank,
	.disable_vblank = smi_disable_vblank,
	.irq_preinstall = smi_irq_preinstall,
	.irq_postinstall = smi_irq_postinstall,
	.irq_uninstall = smi_irq_uninstall,
	.irq_handler = smi_drm_interrupt,
#if RHEL_VERSION_HIGHER_THAN(7, 3)
	.prime_handle_to_fd = drm_gem_prime_handle_to_fd,
	.prime_fd_to_handle = drm_gem_prime_fd_to_handle,

#if RHEL_VERSION_LOWER_THAN(8, 2)
	.gem_prime_import = drm_gem_prime_import,
	.gem_prime_export = drm_gem_prime_export,
	.gem_prime_vmap = smi_gem_prime_vmap,
	.gem_prime_vunmap = smi_gem_prime_vunmap,
	.gem_prime_pin = smi_gem_prime_pin,
	.gem_prime_unpin = smi_gem_prime_unpin,
#elif RHEL_VERSION_LOWER_THAN(8, 3)
	DRM_GEM_VRAM_DRIVER_PRIME,
#endif
#ifdef NEED_PRIME_EXPORT
	// The function is used to export reservation object to secondary GPU.
	// Currently in multi-GPU system, SM768/SM750 only serves as secondary GPU.
	.gem_prime_get_sg_table = smi_gem_prime_get_sg_table,
#endif
	.gem_prime_import_sg_table = smi_gem_prime_import_sg_table,
#endif
};

static int __init smi_init(void)
{
#ifdef CONFIG_VGA_CONSOLE
	if (vgacon_text_force() && smi_modeset == -1)
		return -EINVAL;
#endif

	if (smi_modeset == 0)
		return -EINVAL;
#if RHEL_VERSION_LOWER_THAN(7, 5)
	return drm_pci_init(&driver, &smi_pci_driver);
#else
	return pci_register_driver(&smi_pci_driver);
#endif
}

static void __exit smi_exit(void)
{
#if RHEL_VERSION_LOWER_THAN(7, 5)
	drm_pci_exit(&driver, &smi_pci_driver);
#else
	pci_unregister_driver(&smi_pci_driver);
#endif
}

module_init(smi_init);
module_exit(smi_exit);

MODULE_DEVICE_TABLE(pci, pciidlist);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
